#include "types.hh"
#include "memory.hh"
#include "cpu.hh"

namespace CPU /* Begin module-local namespace */
{
enum { FetchOpcode = 0x7F }; // Anything that evaluates into $Skip.

uint16_t Stack[8], StackHead;  // Internal stack
uint8_t  W_REG;                // Work register
uint16_t NextOpcode = FetchOpcode;
uint16_t NextOpcodeLocation = 0x0000;


/* Here is the whole instruction set defined by the PIC16 architecture: */
/* (W = work register, f = RAM register specified in instruction, k = immediate parameter): */
#define LIST_INSTRUCTIONS(o) \
    /* Basic operations, when writing to RAM registers */ \
    o(MOVWF,  "0000001fffffff", " R    F")   /* f = W */ \
    o(MOVF,   "0010001fffffff", "fT   ZF")   /* f = f,        updates Z      */ \
    o(ANDWF,  "0001011fffffff", "fR  aZF")   /* f = f & W,    updates Z      */ \
    o(IORWF,  "0001001fffffff", "fR  oZF")   /* f = f | W,    updates Z      */ \
    o(XORWF,  "0001101fffffff", "fR  xZF")   /* f = f ^ W,    updates Z      */ \
    o(ADDWF,  "0001111fffffff", "fR  AZF")   /* f = f + W,    updates Z,C,DC */ \
    o(SUBWF,  "0000101fffffff", "fRniAZF")   /* f = f + (-W), updates Z,C,DC the same way as ADDWF */ \
    o(CLRF,   "0000011fffffff", "     ZF")   /* f = 0,        updates Z      */ \
    o(COMF,   "0010011fffffff", "fTn  ZF")   /* f = ~f,       updates Z      */ \
    o(DECF,   "0000111fffffff", "fT d ZF")   /* f = f - 1,    updates Z      */ \
    o(INCF,   "0010101fffffff", "fT i ZF")   /* f = f + 1,    updates Z      */ \
    o(DECFSZ, "0010111fffffff", "fT d SF")   /* f = f - 1, skips next instruction if result zero */ \
    o(INCFSZ, "0011111fffffff", "fT i SF")   /* f = f + 1, skips next instruction if result zero */ \
    o(RRF,    "0011001fffffff", "fT r  F")   /* temp = (C * 0x201 + f*2), f = temp>>2, C = temp&0x002 */ \
    o(RLF,    "0011011fffffff", "fT l  F")   /* temp = (C * 0x201 + f*2), f = temp>>0, C = temp&0x100 */ \
    o(SWAPF,  "0011101fffffff", "fT w  F")   /* f = (f * 0x101) >> 4         */ \
    /* Basic operations, when writing to W register */ \
    o(MOVFw,  "0010000fffffff", "fT   ZW")   /* W = f,        updates Z      */ \
    o(ANDWFw, "0001010fffffff", "fR  aZW")   /* W = f & W,    updates Z      */ \
    o(IORWFw, "0001000fffffff", "fR  oZW")   /* W = f | W,    updates Z      */ \
    o(XORWFw, "0001100fffffff", "fR  xZW")   /* W = f ^ W,    updates Z      */ \
    o(ADDWFw, "0001110fffffff", "fR  AZW")   /* W = f + W,    updates Z,C,DC */ \
    o(SUBWFw, "0000100fffffff", "fRniAZW")   /* W = f + (-W), updates Z,C,DC the same way as ADDWF */ \
    o(CLRW,   "0000010*******", "     ZW")   /* W = 0,        updates Z      */ \
    o(COMFw,  "0010010fffffff", "fTn  ZW")   /* W = ~f,       updates Z      */ \
    o(DECFw,  "0000110fffffff", "fT d ZW")   /* W = f - 1,    updates Z      */ \
    o(INCFw,  "0010100fffffff", "fT i ZW")   /* W = f + 1,    updates Z      */ \
    o(DECFSZw,"0010110fffffff", "fT d SW")   /* W = f - 1, skips next instruction if result zero */ \
    o(INCFSZw,"0011110fffffff", "fT i SW")   /* W = f + 1, skips next instruction if result zero */ \
    o(RRFw,   "0011000fffffff", "fT r  W")   /* temp = (C * 0x201 + f*2), W = temp>>2, C = temp&0x002 */ \
    o(RLFw,   "0011010fffffff", "fT l  W")   /* temp = (C * 0x201 + f*2), W = temp>>0, C = temp&0x100 */ \
    o(SWAPFw, "0011100fffffff", "fT w  W")   /* W = (f * 0x101) >> 4         */ \
    /* Basic operations, with immediate parameter (literal). All of them write into the W register. */ \
    o(MOVLW,  "1100**kkkkkkkk", "IT    W")   /* W = k                        */ \
    o(ANDLW,  "111001kkkkkkkk", "IR  aZW")   /* W = k & W,    updates Z      (bitwise AND)          */ \
    o(IORLW,  "111000kkkkkkkk", "IR  oZW")   /* W = k | W,    updates Z      (bitwise inclusive-OR) */ \
    o(XORLW,  "111010kkkkkkkk", "IR  xZW")   /* W = k ^ W,    updates Z      (bitwise exclusive-OR) */ \
    o(ADDLW,  "11111*kkkkkkkk", "IR  AZW")   /* W = k + W,    updates Z,C,DC (arithmetic add)       */ \
    o(SUBLW,  "11110*kkkkkkkk", "IRniAZW")   /* W = k + (-W), updates Z,C,DC (arithmetic subtract)  */ \
    /* Bit operations (clear,set,test) */ \
    o(BCF,    "0100bbbfffffff", "fbn a F")   /* f = f & ~bit (where bit = 1 << b) */ \
    o(BSF,    "0101bbbfffffff", "fb  o F")   /* f = f |  bit                      */ \
    o(BTFSC,  "0110bbbfffffff", "fb  aS ")   /* skips next instruction if (f & bit) is zero */ \
    o(BTFSS,  "0111bbbfffffff", "fb  aN ")   /* skips next instruction if (f & bit) is NONzero */ \
    /* Special operations */ \
    o(CALL,   "100kkkkkkkkkkk", "  U  G ")   /* push PC+1 into stack; PC = PCLATH,k */ \
    o(GOTO,   "101kkkkkkkkkkk", "     G ")   /*                       PC = PCLATH,k */ \
    o(RETLW,  "1101**kkkkkkkk", "IT   PW")   /* W = k;                pop PC from stack; fetch new instruction */ \
    o(RETURN, "00000000001000", "     P ")   /*                       pop PC from stack; fetch new instruction */ \
    o(RETFIE, "00000000001001", "  g  P ")   /* set INTCON.GIE flag;  pop PC from stack; fetch new instruction */ \
    o(NOP,    "0000000**00000", "       ")   /* NOP is No Operation - do nothing */ \
    /* Unimplemented/dummy operations */ \
    o(OPTION_,"00000001100010", "       ")   /* unimplemented, deprecated instruction */ \
    o(SLEEP,  "00000001100011", "       ")   /* unimplemented */ \
    o(CLRWDT, "00000001100100", "       ")   /* unimplemented */ \
    o(TRIS_,  "000000011001kk", "       ")   /* unimplemented, deprecated instruction */ \
    o($Skip,  "**************", "       ")   /* Catch for any illegal opcodes */

/* The seven-character string describes which actions to perform in order to implement the instruction. */
#define LIST_MICRO_OPERATIONS(o) \
    /* 1st symbol: Load 1st operand (left-hand side) */ \
    o('f', LHS_RAM_REGISTER,   lhs = MemRead(ffield))          /* Load contents of a memory register into LHS */ \
    o('I', LHS_IMMED,          lhs = kfield)                   /* Load literal (immediate) number    into LHS */ \
    /* 2nd symbol: Load 2nd operand (result) */ \
    o('R', RES_W_REGISTER,     res = W_REG)                    /* Load contents of the W register      into RES */ \
    o('T', RES_COPY_LHS,       res = lhs)                      /* Load whatever is in LHS              into RES */ \
    o('b', RES_BITFIELD,       res = 1 << bfield)              /* Load one of: 01,02,04,08,10,20,40,80 into RES */ \
    /* 3rd symbol: Unary updates to the 2nd operand */ \
    o('n', OP_NOT,             res = ~res)                     /* Invert the bits (same as xor by FF) */ \
    o('i', OP_INC,             res = res + 1)                  /* Add 1 */ \
    o('d', OP_DEC,             res = res - 1)                  /* Subtract 1, (same as add FF) */ \
    o('w', OP_SWAP,            res = (res << 4) | (res >> 4))  /* 8-bit rotate by 4 */ \
    o('l', OP_ROTATE_L,        res = (res << 1) | (STATUS.C << 0); STATUS.C = lhs & 0x80) /* 9-bit rotate left by 1 */ \
    o('r', OP_ROTATE_R,        res = (res >> 1) | (STATUS.C << 7); STATUS.C = lhs & 0x01) /* 9-bit rotate right by 1 */ \
    /* 4th symbol: Binary updates to the 2nd operand */ \
    o('a', OP_AND,             res = lhs & res)                /* Bitwise AND of the two operands */ \
    o('o', OP_IOR,             res = lhs | res)                /* Bitwise inclusive OR of the two operands */ \
    o('x', OP_XOR,             res = lhs ^ res)                /* Bitwise exclusive OR of the two operands */ \
    o('A', OP_ADD,             STATUS.C  = ((lhs & 0xFF) + (res & 0xFF)) & 0x100; /* Update carry */ \
                               STATUS.DC = ((lhs & 0xF ) + (res & 0xF )) & 0x10;  /* Update half-carry */ \
                               res = lhs + res)                /* Arithmetically add two operands together */ \
    /* 5th symbol: Possibly act on whether result was zero or not */ \
    o('Z', SET_Z_IF_ZERO,      STATUS.Z = (res == 0))          /* Set Z-flag to indicate whether result was zero */ \
    o('S', SKIP_IF_ZERO,       if(res == 0) DiscardNext())     /* If result was zero, skip next instruction */ \
    o('N', SKIP_IF_NONZERO,    if(res != 0) DiscardNext())     /* If result was nonzero, skip next instruction */ \
    /* 6th symbol: Store the result */ \
    o('F', STORE_RAM_REGISTER, MemWrite(ffield, res))          /* Write into a memory register */ \
    o('W', STORE_W_REGISTER,   W_REG = res)                    /* Write into the W register */ \
    /* 7th symbol: Execution control operations */ \
    o('U', OP_PUSH,            Stack[ StackHead++ & 7 ] = PC.raw - 1)                 /* Save location in stack */ \
    o('P', OP_POP,             PC.raw = Stack[ --StackHead & 7 ]; DiscardNext())      /* Restore location from stack */ \
    o('G', OP_GOTO,            PC.raw = kfield | ((PCLATH & ~7) << 8); DiscardNext()) /* Set location */ \
    o('g', OP_GIE,             INTCON.GIE = 1)                 /* Set global interrupts-enabled flag */

// Note: We don't need OP_NEG. OP_NOT + OP_INC produces the exact same result.
// We also don't need OP_SUB. OP_NOT + OP_INC + OP_ADD produces the exact same result.
// The flag registers after addition and subtraction seem to suggest that
// the PIC internally implements it like this, the same way that the 6502 does.

// It is also worth noting, that the PIC16 architecture could have delay slots for free,
// if they did not NOP the next (already-fetched) instruction that follows a GOTO,
// a CALL or a RETIW/RETURN/RETFIE.
// They must NOP it after a BTFSC/BTFSS (conditional skip), but
// they could choose to not NOP it after any of those other instructions.
// I wonder why they did not go for delay slots. An instruction cycle is wasted for nothing...

/* Enumerate the opcodes (give each one a successive number) */
#define o(mnemonic, bitmask, actions) ind##mnemonic,
enum { LIST_INSTRUCTIONS(o) n_opcodes };
#undef o

/* This array describes all of the opcodes, but the data is only used at compile-time. */
constexpr struct OpcodeDescriptionType
{
    char mnemonic[8]; // Name of the instruction (text). Not used, but could be used for disassembly trace.
    char bitmask[15]; // Encoding of the instruction as an opcode. 14 bits + nul terminator.
    char actions[8];  // Which micro-operations constitute this instruction. 7 symbols + nul terminator.
} instructions[n_opcodes] =
{
    #define o(mnemonic, bitmask, actions) { #mnemonic, bitmask, actions },
    LIST_INSTRUCTIONS(o)
    #undef o
};

/* Create a typedef by the name of each instruction, and a
 * template specialization for returning an instance thereof,
 * given the index.
 * This helps with debugging, as the name of the instruction
 * will be encoded in the function dealing with that instruction.
 * Using the index (the enum from earlier) is necessary because
 * the bitmask search will return an array index position.
 */
template<unsigned index> struct InsType { };
#define o(mnemonic, bitmask, actions) \
    /* Converting typename into an index */ \
    struct mnemonic { enum: unsigned { index = ind##mnemonic }; }; \
    /* Converting index into typename */ \
    template<> struct InsType<ind##mnemonic> { typedef mnemonic result; };
LIST_INSTRUCTIONS(o)
#undef o

/* In a similar line of thought, give a name for each micro-operation,
 * otherwise identified only by their character code (such as 'f' for LHS_RAM_REGISTER). */
template<char c> struct MicroType { typedef void result; };
#define o(character, name, implementation) \
    /* Convert a character into a typename */ \
    template<> struct MicroType<character> { typedef struct name result; };
LIST_MICRO_OPERATIONS(o)
#undef o


/* MatchOpcode() checks - at compile-time - whether the given opcode
 * matches the binary bit pattern given in a string.
 * For example, "01xx0" matches 01000, 01010, 01100 and 01110.
 * The implementation is rather ugly, because C++11 constexpr
 * functions cannot contain loops or variable declarations.
 */
constexpr bool MatchOpcode(unsigned opcode, const char bitmask[], unsigned i=0)
{
    return i >= 14
        || (bitmask[13-i] != char('1' - ((opcode >> i)&1)) && MatchOpcode(opcode,bitmask,i+1));
}

/* FindOpcodeIndex() finds the index into instructions[]
 * for the instruction that is encoded like the parameter.
 *
 * For example:      given the binary number 00010010101101, it returns 3, because
 * instruction 3, which is IORWF, is encoded 0001001fffffff.
 *
 * first_index = first array entry searched.
 * The search is done at compile time.
 */
constexpr unsigned FindOpcodeIndex(unsigned opcode, unsigned first_index=0)
{
    return MatchOpcode(opcode, instructions[first_index].bitmask)
           ? first_index
           : FindOpcodeIndex(opcode, first_index+1);
}

/* ReadField() returns information on how to read a field from an opcode,
 * indicated by bitmask[] and whichfield. (At compile time of course!)
 *
 * Return value is: firstbit = Position of bitmask from right = min(foundpositions)
 *                  num_bits = Number of bits in the bitmask  = count(foundpositions)
 *
 * For example, for bitmask="0010bbbfffffff", whichfield='b', it returns {7,3}.
 */
struct FieldResult { unsigned firstbit, num_bits; };
namespace std { template<typename T> constexpr const T& min(const T& a, const T& b) { return a<b ? a : b; } }
constexpr FieldResult ReadField(const char bitmask[], char whichfield,
                                unsigned i=0, FieldResult res={14,0})
{
    return i >= 14
        ? res
        : bitmask[13-i] == whichfield
            ? ReadField(bitmask,whichfield, i+1, { std::min(i, res.firstbit), res.num_bits+1 } )
            : ReadField(bitmask,whichfield, i+1, res);
}

/* MicroExecutor executes all possible instructions in a series of micro-operations. */
struct MicroExecutor
{
    unsigned kfield, bfield, ffield; // respective parts of the opcode
    uint8_t  lhs, res;               // input and outcome

    /* Execute one micro-operation (part of the instruction) */
    template<char character, typename T = typename MicroType<character>::result>
    void Execute()
    {
        #define o(ch, name, implementation) if(ch == character) { implementation; }
        LIST_MICRO_OPERATIONS(o)
        #undef o
    }
};

/* ExecuteOpcode() executes the given instruction.
 * The instruction is given as a mnemonic type rather than an index.
 * This way, you will get e.g. ExecuteOpcode<ADDWF> rather than ExecuteOpcode<5>
 * in the symbol listing and in backtraces. Makes debugging so much easier!
 */
template<typename mnemonic, unsigned ind = mnemonic::index>
void ExecuteOpcode(unsigned opcode)
{
    constexpr auto kfield = ReadField(instructions[ind].bitmask, 'k');
    constexpr auto bfield = ReadField(instructions[ind].bitmask, 'b');
    constexpr auto ffield = ReadField(instructions[ind].bitmask, 'f');

    MicroExecutor e { (opcode >> kfield.firstbit) & ((1 << kfield.num_bits) - 1),
                      (opcode >> bfield.firstbit) & ((1 << bfield.num_bits) - 1),
                      (opcode >> ffield.firstbit) & ((1 << ffield.num_bits) - 1),
                      0,0
                    };

    // Execute in sequence all micro-operations constituting this instruction.
    e.Execute<instructions[ind].actions[0]>();
    e.Execute<instructions[ind].actions[1]>(); // a for-loop is impossible here,
    e.Execute<instructions[ind].actions[2]>(); // because these are template-parameters.
    e.Execute<instructions[ind].actions[3]>();
    e.Execute<instructions[ind].actions[4]>();
    e.Execute<instructions[ind].actions[5]>();
    e.Execute<instructions[ind].actions[6]>();
}


/* All opcodes in form 0000000nnnnnnn: */
static void (*ExecuteLo7[0x80])(unsigned) =
{
    /* Make a pointer to the correct template function for opcode n */
    #define qq(n)   ExecuteOpcode<InsType<FindOpcodeIndex(n)>::result>,
    #define q(n)    qq(n) qq(n+0x01)
    /* Make a table of 128 of them. */
    q(0x00) q(0x02) q(0x04) q(0x06) q(0x08) q(0x0A) q(0x0C) q(0x0E)
    q(0x10) q(0x12) q(0x14) q(0x16) q(0x18) q(0x1A) q(0x1C) q(0x1E)
    q(0x20) q(0x22) q(0x24) q(0x26) q(0x28) q(0x2A) q(0x2C) q(0x2E)
    q(0x30) q(0x32) q(0x34) q(0x36) q(0x38) q(0x3A) q(0x3C) q(0x3E)
    q(0x40) q(0x42) q(0x44) q(0x46) q(0x48) q(0x4A) q(0x4C) q(0x4E)
    q(0x50) q(0x52) q(0x54) q(0x56) q(0x58) q(0x5A) q(0x5C) q(0x5E)
    q(0x60) q(0x62) q(0x64) q(0x66) q(0x68) q(0x6A) q(0x6C) q(0x6E)
    q(0x70) q(0x72) q(0x74) q(0x76) q(0x78) q(0x7A) q(0x7C) q(0x7E)
    #undef q
};

void ExecuteOpcode00(unsigned opcode)
{
    ExecuteLo7[opcode] (opcode);
}

/* All opcodes in form nnnnnnnxxxxxxx where x=any bit: */
static void (*ExecuteHi7[0x80])(unsigned) =
{
    /* Make a pointer to the correct template function for opcode n*0x80 */
    #define pp(n)   ExecuteOpcode<InsType<FindOpcodeIndex((n)*0x80)>::result>,
    #define q(n)    pp(n) pp(n+0x01)
    /* Make a table of 128 of them. The first must be ExecuteOpcode00. */
    ExecuteOpcode00,
    pp(0x01)q(0x02) q(0x04) q(0x06) q(0x08) q(0x0A) q(0x0C) q(0x0E)
    q(0x10) q(0x12) q(0x14) q(0x16) q(0x18) q(0x1A) q(0x1C) q(0x1E)
    q(0x20) q(0x22) q(0x24) q(0x26) q(0x28) q(0x2A) q(0x2C) q(0x2E)
    q(0x30) q(0x32) q(0x34) q(0x36) q(0x38) q(0x3A) q(0x3C) q(0x3E)
    q(0x40) q(0x42) q(0x44) q(0x46) q(0x48) q(0x4A) q(0x4C) q(0x4E)
    q(0x50) q(0x52) q(0x54) q(0x56) q(0x58) q(0x5A) q(0x5C) q(0x5E)
    q(0x60) q(0x62) q(0x64) q(0x66) q(0x68) q(0x6A) q(0x6C) q(0x6E)
    q(0x70) q(0x72) q(0x74) q(0x76) q(0x78) q(0x7A) q(0x7C) q(0x7E)
    #undef q
};

/* Technically, we could make a table of 16384 ExecuteOpcode pointers
 * in the same way we just made a table of 128 of them, but it would be
 * bad for cache efficiency in the emulation.
 */

} /* End module-local namespace */

void ExecuteOpcode()
{
    unsigned CurrentOpcode = CPU::NextOpcode;

    // All instructions begin with fetching the next opcode
    CPU::NextOpcodeLocation = PC.raw++ % count(ProgramMemory);
    CPU::NextOpcode         = ProgramMemory[CPU::NextOpcodeLocation];

    CPU::ExecuteHi7[CurrentOpcode >> 7] (CurrentOpcode);
}

void DiscardNext()
{
    /* If the next opcode was invalidated by a GOTO, RETURN,
     * RETLW, RETFIE, INCFSZ, DECFSZ, BTFSC or BTFSS, change
     * the already fetched opcode into an internal NOP opcode.
     * This causes the opcode in the next location to be fetched
     * at the next instruction, and an extra cycle being spent there.
     * This is also called from memory.cc upon write into PCL.
     */
    CPU::NextOpcode = CPU::FetchOpcode;
}






























