#include <cstdio>
#include <memory>

#include "memory.hh"

/* Parse an Intel IHEX file. These files typically store data to be written to a chip.
 * The format is described here: http://en.wikipedia.org/wiki/Intel_HEX
 */
void LoadHexFile(const char* fn)
{
    /* Open file */
    struct fcloser { void operator() (std::FILE* fp) { if(fp) std::fclose(fp); } };
    /* The std::unique_ptr ensures that the file is automatically closed, */
    /* no matter which path this function is exited.                      */
    std::unique_ptr<std::FILE, fcloser> fp { std::fopen(fn, "rb") };

    if(!fp)
    {
        std::perror(fn);
        return;
    }

    /* Read each line */
    for(unsigned temp, linelen, ext_address = 0, lineno = 1; ; ++lineno)
    {
        /* Read a line from the file */
        char Buf[522], checksum=0x00;
        if(!std::fgets(Buf, sizeof(Buf), fp.get())) break;

        /* Convert the line into a hexadecimal stream */
        auto* Byte = reinterpret_cast<unsigned char*>(&Buf[0]); // Reuse the same buffer.
        for(linelen=0; Buf[linelen]!='\r' && Buf[linelen]!='\n' && Buf[linelen]; ++linelen)
        {
            // Read a two-digit hexadecimal number from the line and store it into the buffer.
            // The first character (which should be ':') is skipped.
            if(linelen&1)    { std::sscanf(Buf+linelen, "%02X", &temp);  }
            else if(linelen) { Byte[linelen/2] = temp; checksum += temp; }
        }

        /* Check whether the line was valid */
        unsigned length = Byte[1]; // First byte indicates the line length.
        bool valid = Buf[0] == ':' && linelen == 11+length*2 && checksum==0;

        /* Interpret the line (indicated by 4th byte). */
        switch(valid ? Byte[4] : 0xFF)
        {
            case 0: // Data record
                /* The data is word-oriented (on the PIC), but IHEX is a byte-by-byte format. */
                #define PutByteIn(where) do { \
                    if(addr&1) where = (where&0x00FF) + 0x100 * Byte[5+a]; \
                    else       where = (where&0xFF00) + 0x001 * Byte[5+a]; } while(0)
                for(unsigned addr = Byte[2]*0x100u + Byte[3] + ext_address, a=0; a<length; ++a, ++addr)
                {
                    /* PIC-specific: Use the address to determine which memory should be programmed */
                    if(addr < count(ProgramMemory)*2) PutByteIn(ProgramMemory[ addr/2 ]);
                    else if(addr >= 0x2100*2)         PutByteIn(EEPROM[ (addr/2) % count(EEPROM) ]);
                    /* For now, we are ignoring the configuration word (at 0x2007). */
                }
                break;

            case 1: // End of hex file
                break;

            case 4: // Extended address
                ext_address = (Byte[5]*0x100u + Byte[6]) * 0x8000u;
                break;

            default: // Unknown record type or an invalid file
                fprintf(stderr, "Invalid line %u in IHEX file "
                                "(begin '%c',length %u,checksum %u,record type %02X)expect(':',%u,0)\n",
                        lineno, Buf[0], length, checksum&0xFF, Byte[4], (linelen-11)/2);
        }
    }
}



