#include <type_traits>

#include "memory.hh"
#include "cpu.hh"

uint16_t ProgramMemory[2048];
uint8_t EEPROM[128];
uint8_t RAM[224];  // In three banks, of sizes 0x50+0x10, 0x50, 0x30.

unsigned long long CycleCounter=0;

// Yes, we already saw this list once, but it was in a header file.
// This is the actual definition. This redundancy (repetition of information)
// is why I don't like to create multi-file programs for my videos.
// Nobody likes to read this boring stuff twice when watching a video.
PCtype      PC;
STATUStype  STATUS;
OPTIONtype  OPTION;
T2CONtype   T2CON;
INTCONtype  INTCON;
CCP1CONtype CCP1CON;
uint8_t     PCLATH=0x00, FSR, EEADR, EEDATA, EECON1=0x00, EECON2;
uint8_t     PR2=0xFF, TMR0, CCPR1L, CCPR1H, TMR2=0x00;
uint16_t    TMR0_prescaler=0, TMR2_prescaler=0;

/* Compile-time table of memory ranges */
static constexpr struct MemoryRangeType
{
    uint8_t* Data; // Interestingly enough, a pointer to non-constexpr data is a valid constexpr.
    struct Range
    {
        unsigned minimum = ~0u, maximum = ~0u;
        constexpr Range() {}
        constexpr Range(unsigned one)           : minimum{one},maximum{one+1} {}
        constexpr Range(unsigned mi,unsigned ma): minimum{mi},maximum{ma+1}   {}
    } Ranges[4];
} MemoryRanges[] =
{
    // List of memory areas that are RAM. Eg. 0x20-0x7F is mapped into RAM[0x00-0x5F].
    { &RAM[0x00],  {{0x20,0x7F}}                             }, // 80 bytes of RAM on page 0 (+16 on common)
    { &RAM[0x50],  {{0xF0,0xFF},{0x170,0x17F},{0x1F0,0x1FF}} }, // 16 bytes of RAM (common on all pages)
    { &RAM[0x60],  {{0xA0,0xEF}}                             }, // 80 bytes of RAM on page 1
    { &RAM[0xB0],  {{0x120,0x14F}}                           }, // 48 bytes of RAM on page 2
    // List of supported I/O registers that do not have any on-write side effects.
    // Note: We do not emulate NEARLY the whole set of peripherals that PIC has to offer,
    //       and we have severely cut corners even in the parts that we do emulate.
    { &TMR0,       {{0x01},{0x101}}                }, // TMR0 can be accessed in addresses 0x01 and 0x101.
    { &STATUS.raw, {{0x03},{0x83},{0x103},{0x183}} },
    { &FSR,        {{0x04},{0x84},{0x104},{0x184}} },
    { &PCLATH,     {{0x0A},{0x8A},{0x10A},{0x18A}} },
    { &INTCON.raw, {{0x0B},{0x8B},{0x10B},{0x18B}} },
    { &TMR2,       {{0x11}}                        },
    { &T2CON.raw,  {{0x12}}                        },
    { &CCPR1L,     {{0x15}}                        },
    { &CCPR1H,     {{0x16}}                        },
    { &CCP1CON.raw,{{0x17}}                        },
    { &OPTION.raw, {{0x81},{0x181}}                },
    { &PR2,        {{0x92}}                        },
    { &EEDATA,     {{0x9A}}                        },
    { &EEADR,      {{0x9B}}                        },
    // These are implemented separately because of side-effects on read/write:
    //  $indirect$ at 00, 80, 100, 180
    //  PCL        at 02, 82, 102, 182
    //  EECON1     at 9C
    //  EECON2     at 9D
    // Missing (not implemented):
    //  PORT A:  PORTA @05;     TRISA @85;        CMCON @1F; VRCON @9F
    //  PORT B:  PORTB @06,106; TRISB @86,186
    //  TIMER1:  TMR1L @0E; TMR1H @0F; T1CON @10
    //  USART:   RCSTA @18; TXREG @19; RCREG @1A; TXSTA @9A; SPBRG @99; PIR1 @0C; PIE1 @8C
};

// A helper function for finding what is in the given memory address. At compile-time!
static constexpr uint8_t* FindMemory(unsigned Addr, unsigned begin=0, unsigned n=0,
    // Because of the silly C++11 requirement that constexpr function body be a return statement,
    // some constants are declared here as default parameters rather than within the function body.
    const unsigned n_ranges = count(MemoryRanges[0].Ranges))
{
    return
        // Are we within the memory element table bounds?
        begin < count(MemoryRanges)
        // Address within the memory range?
        ? (Addr >= MemoryRanges[begin].Ranges[n].minimum
        && Addr <  MemoryRanges[begin].Ranges[n].maximum)
            // Return a pointer to that memory item (offset by memory range begin)
            ? MemoryRanges[begin].Data + (Addr - MemoryRanges[begin].Ranges[n].minimum)
            // Otherwise, try next memory range. After 4 ranges, next memory element.
            : FindMemory(Addr, begin + (n+1)/n_ranges, (n+1)%n_ranges)
        // No more memory element table items. Thus, nothing mapped here.
        : nullptr;
}

// This is the actual memory access function.
static uint8_t MemAccess9(uint16_t address, bool write, uint8_t value=0)
{
    // Using FindMemory(), populate a lookup table of 0x200 memory pointers.
    #define f FindMemory
    #define f8(n) f(n), f(n+1), f(n+2), f(n+3), f(n+4), f(n+5), f(n+6), f(n+7)
    #define f64(n) f8(n),f8(n+8),f8(n+16),f8(n+24),f8(n+32),f8(n+40),f8(n+48),f8(n+56)
    static constexpr uint8_t* MemoryMap[0x200]
    { f64(0x000),f64(0x040),f64(0x080),f64(0x0C0),
      f64(0x100),f64(0x140),f64(0x180),f64(0x1C0) };
    #undef f8
    #undef f64
    #undef f

    // Is there a trivial I/O port or RAM in this memory address? If so, access it.
    if(MemoryMap[address] != nullptr)
    {
        uint8_t& r = *MemoryMap[address];
        if(write) r = value;
        return r;
    }

    // Is it one of those I/O ports that have side effects?
    switch(address)
    {
        // Indirect memory access
        case 0x00: case 0x80: case 0x100: case 0x180:
            if(!(FSR & 0x7F)) return 0x00; // Avoid infinite recursion
            return MemAccess9( FSR | (STATUS.IRP << 8), write, value );

        // PCL (program counter low 8 bits)
        case 0x02: case 0x82: case 0x102: case 0x182:
            if(write)
            {
                PC.PCH = PCLATH; // PCH is automatically updated from PCLATH
                PC.PCL = value;
                // The program pointer was changed.
                // Induce a 1 cycle delay, force fetch of new opcode.
                DiscardNext();
            }
            return PC.PCL;

        // EECON1 (EEPROM control register 1)
        case 0x9C:
            if(write)
            {
                // EEPROM read and write
                // Check if RD flag was set
                if((value & 0x01))
                    { EEDATA = EEPROM[EEADR % count(EEPROM)]; }
                // Check if WR flag was set
                if((value & 0x02) && (EECON2 == 0xFF) && (EECON1 & 0x08))
                    { EEPROM[EEADR % count(EEPROM)] = EEDATA; EECON2 = 0x00; }
                // Clear any bit except WREN
                EECON1 = value & 0x08;
            }
            return EECON1;

        // EECON2 (EEPROM control register 2)
        case 0x9D:
            // This is technically wrong implementation, but as said before,
            // I am cutting corners to keep the video length in control.
            if(write) EECON2 |= value;
            break;
    }

    // Anything else: Unimplemented
    return 0x00;
}

uint8_t MemRead(uint8_t addr)
{
    // The input is a 7-bit address. Make it a 9-bit.
    return MemAccess9( addr | (STATUS.RP << 7), false );
}

void MemWrite(uint8_t addr, uint8_t value)
{
    // The input is a 7-bit address. Make it a 9-bit.
    MemAccess9( addr | (STATUS.RP << 7), true, value );
}




