#ifndef bqtPicMemoryHH
#define bqtPicMemoryHH

#include <stdint.h>

#include "types.hh"

/* This chip has 2048 bytes of read-only program memory. These bytes are 14-bit bytes. */
extern uint16_t ProgramMemory[2048];

/* Then there is 224 bytes of 8-bit RAM. They are also called general-purpose registers. */
extern uint8_t RAM[224];

/* Finally, 128 bytes of non-volatile memory (RAM). It preserves contents during power-off. */
extern uint8_t EEPROM[128];


/* This macro makes declaring the special registers below so much more compact. */
#define MakeRegister(inttype,name, ...) \
    extern union name##type \
    { \
        template<unsigned bitposition, unsigned num_bits = 1> \
            using R = RegBitSet<bitposition, num_bits, inttype>; \
        __VA_ARGS__; \
        inttype raw; \
    } name

/* Declare a number of special registers that contain bit fields.      */
/* The PIC16F628A contains a heap of special registers, each of which  */
/* have plenty of bits. In this emulator, we only emulate a very small */
/* subset of the available features.                                   */
/* It would be injustice to even attempt to describe the purpose and   */
/* function of each register here. That's what the datasheet is for.   */

/* Program counter. It has a low part and a high part. */
MakeRegister(uint16_t,PC,      R<0,8>PCL;      R<8,8>PCH);

/* Status/flags register. C=Carry, DC=half-carry, Z=zero, RP&IRP = page select bits. */
MakeRegister(uint8_t, STATUS,  R<0>C; R<1>DC;  R<2>Z; R<5,2>RP; R<7>IRP);

/* The only bits we care about in OPTION are those that control TIMER0 speed. */
MakeRegister(uint8_t, OPTION,  R<0,3>PreScale; R<3>PSA);

/* The only bits we care about in T2CON are those that control TIMER2 speed. */
MakeRegister(uint8_t, T2CON,   R<0,2>PreScale; R<2>Enabled);

/* In the INTCON register, we only care about T0IF. This bit tells whether */
/* the counter cumulated by TIMER0 has wrapped. Additionally, we need the  */
/* GIE bit, because the RETFIE instruction of the CPU can reset that bit.  */
MakeRegister(uint8_t, INTCON,  R<2>T0IF;       R<7>GIE);

/* The CCP1CON register controls how the PWM operation (TIMER2) works. */
MakeRegister(uint8_t, CCP1CON, R<0,4>mode;     R<4,2>lsb);


/* More registers! The PIC has a heap of registers indeed! This is why  */
/* it took such a long time for me to make this video. I wasn't sure    */
/* how I am ever going to be able to explain all the stuff that I type. */
/* Just look at the datasheet if you want to know what they are.        */
extern uint8_t  PCLATH, FSR, EEADR, EEDATA, EECON1, EECON2;
extern uint8_t  PR2, TMR0, CCPR1L, CCPR1H, TMR2;
extern uint16_t TMR0_prescaler, TMR2_prescaler;


/* Accessors for memory reads and writes. The address parameter is 7-bit.            */
/* The whole memory address is 9-bit; the two missing bits are taken from STATUS.RP. */
extern uint8_t MemRead(uint8_t address);
extern void MemWrite(uint8_t address, uint8_t value);


extern unsigned long long CycleCounter;

#endif



